import re

import pytest

from sdmx import to_sdmx
from sdmx.reader import (
    detect_content_reader,
    get_reader_for_media_type,
    get_reader_for_path,
    json,
)
from sdmx.reader.base import BaseReader, Converter


class TestConverter:
    def test_handles(self) -> None:
        assert False is Converter.handles({}, {})

    def test_convert(self) -> None:
        with pytest.raises(NotImplementedError):
            Converter().convert({})


class TestBaseReader:
    def test_deprecated_read_message(self) -> None:
        with (
            pytest.raises(NotImplementedError),
            pytest.warns(DeprecationWarning, match="use Converter.convert"),
        ):
            BaseReader().read_message({})


def test_deprecated_detect_content_reader() -> None:
    with pytest.warns(DeprecationWarning, match=r"use get_reader\(\) instead"):
        assert json.Reader is detect_content_reader(b"{")


@pytest.mark.parametrize(
    "value",
    [
        "application/x-pdf",
    ],
)
def test_deprecated_get_reader_for_media_type0(value):
    with (
        pytest.raises(
            ValueError, match=re.escape(f"Media type {value!r} not supported by any of")
        ),
        pytest.warns(DeprecationWarning, match=r"use get_reader\(requests.Response"),
    ):
        get_reader_for_media_type(value)


@pytest.mark.parametrize(
    "value",
    [
        "application/vnd.sdmx.data+xml; version=3.0.0",
        "application/xml;charset=UTF-8",
        "draft-sdmx-json;charset=UTF-8",
    ],
)
def test_deprecated_get_reader_for_media_type1(value):
    with pytest.warns(DeprecationWarning, match=r"use get_reader\(requests.Response"):
        # Does not raise
        get_reader_for_media_type(value)


@pytest.mark.parametrize(
    "path",
    (
        "example.json",
        pytest.param("example.badsuffix", marks=pytest.mark.xfail(raises=ValueError)),
    ),
)
def test_deprecated_get_reader_for_path(path):
    with pytest.warns(DeprecationWarning, match=r"use get_reader\(\) instead"):
        assert json.Reader is get_reader_for_path(path)


def test_to_sdmx():
    with pytest.raises(NotImplementedError, match="Convert <class 'dict"):
        to_sdmx(dict())



import pytest

import sdmx
from sdmx.reader.json import Reader


class TestReader:
    def test_deprecated_detect(self) -> None:
        with pytest.warns(DeprecationWarning, match="use Converter.handles"):
            assert True is Reader.detect(b"{")


@pytest.mark.parametrize_specimens("path", format="json")
def test_json_read(path):
    """Test that the samples from the SDMX-JSON spec can be read."""
    sdmx.read_sdmx(path)


@pytest.mark.xfail(reason="Test some code for the incomplete SDMX-JSON 2.0 reader")
def test_json_read_v2(specimen):
    sdmx.read_sdmx(
        specimen.base_path.joinpath("v3", "json", "structure", "generated-sample.json")
    )


def test_header(specimen):
    with specimen("flat.json") as f:
        resp = sdmx.read_sdmx(f)
    assert resp.header.id == "62b5f19d-f1c9-495d-8446-a3661ed24753"



from pathlib import Path

import sdmx
from sdmx.model import v21 as model
from sdmx.testing import MessageTest


class Test_ESTAT_dsd_apro_mk_cola(MessageTest):
    directory = "ESTAT"
    filename = "apro_mk_cola-structure.xml"

    def test_codelists_keys(self, msg):
        assert len(msg.codelist) == 6
        assert isinstance(msg.codelist.CL_GEO, model.Codelist)

    def test_codelist_name(self, msg):
        assert msg.codelist.CL_GEO.UK.name.en == "United Kingdom"
        assert msg.codelist.CL_FREQ.name.en == "FREQ"

    def test_code_cls(self, msg):
        assert isinstance(msg.codelist.CL_FREQ.D, model.Code)

    def test_writer(self, msg):
        cls_as_dfs = sdmx.to_pandas(msg.codelist)

        # Number of codes expected in each Codelist
        count = {
            "CL_FREQ": 6,
            "CL_GEO": 41,
            "CL_OBS_FLAG": 10,
            "CL_OBS_STATUS": 3,
            "CL_PRODMILK": 12,
            "CL_UNIT": 1,
        }

        assert all(len(df) == count[id] for id, df in cls_as_dfs.items())

    def test_urn(self, msg):
        """https://github.com/dr-leo/pandaSDMX/issue/154."""
        expected = (
            "urn:sdmx:org.sdmx.infomodel.datastructure.DataStructure="
            "ESTAT:DSD_apro_mk_cola(1.0)"
        )

        # URN is parsed
        assert msg.structure.DSD_apro_mk_cola.urn == expected


class TestDSDCommon(MessageTest):
    directory = "SGR"
    filename = "common-structure.xml"

    def test_codelists_keys(self, msg):
        assert len(msg.codelist) == 5
        assert isinstance(msg.codelist.CL_FREQ, model.Codelist)

    def test_codelist_name(self, msg):
        assert msg.codelist.CL_FREQ.D.name.en == "Daily"

    def test_code_cls(self, msg):
        assert isinstance(msg.codelist.CL_FREQ.D, model.Code)

    def test_annotations(self, msg):
        code = msg.codelist.CL_FREQ.A
        anno_list = list(code.annotations)
        assert len(anno_list) == 1
        a = anno_list[0]
        assert isinstance(a, model.Annotation)
        assert a.text.en.startswith("It is")
        assert a.type == "NOTE"


class TestECB_EXR1(MessageTest):
    directory = Path("ECB_EXR", "1")
    filename = "structure.xml"

    def test_uri(self, msg):
        """https://github.com/dr-leo/pandaSDMX/issue/154."""
        expected = "https://www.ecb.europa.eu/vocabulary/stats/exr/1"
        # URI is parsed
        assert msg.structure["ECB_EXR1"].uri == expected


def test_exr_constraints(specimen):
    with specimen("1/structure-full.xml") as f:
        msg = sdmx.read_sdmx(f)
    ECB_EXR1 = msg.structure["ECB_EXR1"]

    # Test DimensionDescriptor
    dd = ECB_EXR1.dimensions

    # Correct order
    assert dd[0].id == "FREQ"

    # Correct number of dimensions
    assert len(dd.components) == 6

    # Dimensions can be retrieved by name; membership can be tested
    assert "W" in dd.get("FREQ")

    # Similar tests for AttributeDescriptor
    ad = ECB_EXR1.attributes
    assert len(ad.components) == 24
    assert ad[-1].id == "UNIT_MULT"
    assert "5" in ad.get("UNIT_MULT")

    cc = msg.constraint["EXR_CONSTRAINTS"]

    # Commented: very slow
    # # Expected number of constrained keys
    # keys = list(cc.iter_keys(ECB_EXR1))
    # assert 5 * 58 * 59 * 12 * 6 * 1 == len(keys)

    del cc

    # TODO update the following for a complete implementation

    # assert "W" not in m._constrained_codes.FREQ
    #
    # key = {"FREQ": ["W"]}
    #
    # assert m.in_codes(key)
    #
    # assert not m.in_constraints(key, raise_error=False)
    #
    # with pytest.raises(ValueError):
    #     m.in_constraints(key)
    #
    # assert m.in_constraints({"CURRENCY": ["CHF"]})
    #
    # # test with invalid key
    # with pytest.raises(TypeError):
    #     m._in_constraints({"FREQ": "A"})
    #
    # # structure writer with constraints
    # out = sdmx.to_pandas(m)
    # cl = out.codelist
    # assert cl.shape == (3555, 2)
    #
    # # unconstrained codelists
    # out = sdmx.to_pandas(m, constraint=False)
    # cl = out.codelist
    # assert cl.shape, (4177, 2)



from . import Source as BaseSource


class Source(BaseSource):
    _id = "INSEE"

    def modify_request_args(self, kwargs):
        """Supply explicit provider agency ID for INSEE.

        This web service accepts either "ALL" or "FR1" as a provider agency ID for
        structure endpoints, but not "INSEE" (see :issue:`21`).

        This hook sets the provider to "ALL" for structure queries if it is not given
        explicitly.
        """
        super().modify_request_args(kwargs)

        # NB this is an indirect test for resource_type != 'data'; because of
        #    the way the hook is called, resource_type is not available
        #    directly.
        if "key" not in kwargs:
            kwargs.setdefault("provider", "ALL")



import pytest

from sdmx.format import Flag, MediaType
from sdmx.model import v21
from sdmx.reader.base import BaseReader
from sdmx.reader.xml.v21 import Reader as XMLv21Reader


class TestBaseReader:
    @pytest.fixture
    def MinimalReader(self):
        """A reader that implements the minimum abstract methods."""

        class cls(BaseReader):
            media_types = [
                MediaType("", "xml", "2.1", Flag.data, full="application/foo"),
            ]

            def convert(self, source, dsd=None):
                pass  # pragma: no cover

        return cls

    def test_deprecated_kwarg(self):
        r = XMLv21Reader()

        dsd0 = v21.DataStructureDefinition(id="FOO")
        dsd1 = v21.DataStructureDefinition(id="BAR")
        with (
            pytest.warns(
                DeprecationWarning, match="dsd=.* keyword argument; use structure="
            ),
            pytest.raises(ValueError, match="Mismatched structure=FOO, dsd=BAR"),
        ):
            r.convert(None, structure=dsd0, dsd=dsd1)

    def test_detect(self, MinimalReader):
        with pytest.warns(DeprecationWarning, match="use Converter.handles"):
            assert False is MinimalReader.detect(b"foo")

    def test_handles_media_type(self, caplog, MinimalReader):
        """:meth:`.handles_media_type` matches even when params differ, but logs."""
        with pytest.warns(DeprecationWarning, match="use Converter.handles"):
            assert True is MinimalReader.handles_media_type("application/foo; bar=qux")
        assert (
            "Match application/foo with params {'bar': 'qux'}; "
            "expected {'version': '2.1'}" in caplog.messages
        )



import io
import logging
import pathlib
from functools import lru_cache
from typing import TYPE_CHECKING, Any, ClassVar, Optional
from warnings import warn

import requests

from sdmx.format import MediaType

if TYPE_CHECKING:
    import sdmx.message
    import sdmx.model.common

log = logging.getLogger(__name__)


class Converter:
    """Base class for conversion to :mod:`sdmx` objects."""

    @classmethod
    def handles(cls, data: Any, kwargs: dict) -> bool:
        """Return :any:`True` if the class can convert `data` using `kwargs`."""
        return False

    def convert(self, data: Any, **kwargs) -> "sdmx.message.Message":
        """Convert `data` to an instance of an SDMX Message subclass."""
        raise NotImplementedError


class BaseReader(Converter):
    """Converter of file or binary data in standard SDMX formats."""

    #: First byte(s) of file or response body content, used by
    #: :meth:`~.BaseReader.handles`.
    binary_content_startswith: ClassVar[Optional[bytes]] = None

    #: List of media types, used by :meth:`.handles`.
    media_types: ClassVar[list[MediaType]] = []

    #: List of file name suffixes, used by :meth:`.handles`.
    suffixes: ClassVar[list[str]] = []

    @classmethod
    def detect(cls, content: bytes) -> bool:
        """Detect whether the reader can handle `content`.

        .. deprecated:: 2.20.0
           Use :meth:`~.BaseReader.handles` instead.

        Returns
        -------
        bool
            :obj:`True` if the reader can handle the content.
        """
        warn(
            "BaseReader.detect(bytes); use Converter.handles() instead",
            DeprecationWarning,
        )
        return False

    @classmethod
    @lru_cache()
    def handles_media_type(cls, value: str) -> bool:
        """:obj:`True` if the reader can handle content/media type `value`.

        .. deprecated:: 2.20.0
           Use :meth:`~.BaseReader.handles` instead.
        """
        warn(
            "BaseReader.handles_media_type(str); use Converter.handles(requests.Response) instead",
            DeprecationWarning,
        )
        for mt in cls.media_types:
            if mt.match(value):
                return True
        return False

    @classmethod
    def supports_suffix(cls, value: str) -> bool:
        """:obj:`True` if the reader can handle files with suffix `value`.

        .. deprecated:: 2.20.0
           Use :meth:`~.BaseReader.handles` instead.
        """
        warn(
            "BaseReader.supports_suffix(str); use Converter.handles(pathlib.Path) instead",
            DeprecationWarning,
        )
        return cls.handles(pathlib.Path(f"_.{value.lower()}"), {})

    @classmethod
    def handles(cls, data, kwargs):
        """Return :any:`True` if the Reader can convert `data` using `kwargs`.

        The default implementation checks for any of the following conditions:

        1. `data` is :class:`pathlib.Path` and has one of the Reader's
           :attr:`.suffixes`. The match is case-insensitive.
        2. `data` is :class:`requests.Response` and its
           :attr:`~requests.Response.headers` include a ``content-type`` that is matched
           by one of the :class:`.MediaTypes` in the Reader' :attr:`.media_types`.
        3. `data` is :class:`bytes`, :class:`io.IOBase`, or :class:`io.BufferedReader`
           and starts with the class' :attr:`.binary_content_startswith` (if any). For
           the :mod:`io` classes, this check is performed by 'peeking' at the content
           without changing the position in the file for a later call to
           :meth:`.convert`.
        """
        if isinstance(data, pathlib.Path):
            # `data` is a Path with a known suffix
            # Formerly supports_suffix()
            return data.suffix.lower() in cls.suffixes
        elif isinstance(data, requests.Response):
            # `data` is a HTTP Response with given content-type headers
            # Formerly handles_media_type()
            value = data.headers.get("content-type", "")
            return any(mt.match(value) for mt in cls.media_types)
        elif bcsw := cls.binary_content_startswith:
            if isinstance(data, bytes):
                # `data` is raw bytes
                # Formerly detect()
                peek = data
            elif isinstance(data, io.BufferedReader):
                # `data` is a subtype of io.IOBase that supports peek()
                peek = data.peek(len(bcsw))
            elif isinstance(data, io.IOBase):
                # `data` is a subtype of io.IOBase that supports tell()/seek()
                # Formerly in read_sdmx()
                pos = data.tell()
                peek = data.readline().strip()
                data.seek(pos)  # Return to original `pos`ition
            else:
                peek = b""

            return peek.startswith(bcsw)

        return False

    def convert(
        self, data, structure: Optional["sdmx.model.common.Structure"] = None, **kwargs
    ):
        """Convert `data` to an instance of an SDMX Message subclass.

        Parameters
        ----------
        data : file-like
            Message content.
        structure :
            :class:`DataStructure <.BaseDataStructureDefinition>` or
            :class:`MetadataStructure <.BaseMetadataStructureDefinition>`
            for aid in reading `source`.

        Returns
        -------
        :class:`.Message`
            An instance of a Message subclass.
        """
        raise NotImplementedError

    def read_message(self, *args, **kwargs):
        """Deprecated. Use :meth:`.convert` instead."""
        warn(
            "Reader.read_message(); use Converter.convert() instead",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.convert(*args, **kwargs)

    @classmethod
    def _handle_deprecated_kwarg(
        cls, structure: Optional["sdmx.model.common.Structure"], kwargs
    ) -> Optional["sdmx.model.common.Structure"]:
        try:
            dsd = kwargs.pop("dsd")
        except KeyError:
            dsd = None
        else:
            warn(
                "Reader.read_message(…, dsd=…) keyword argument; use structure=…",
                DeprecationWarning,
                stacklevel=2,
            )
            if structure and structure is not dsd:
                raise ValueError(f"Mismatched structure={structure}, dsd={dsd}")
        return structure or dsd



import operator

import pytest
from packaging.version import InvalidVersion
from packaging.version import Version as PVVersion

from sdmx.model.version import Version, increment, parse


class TestVersion:
    @pytest.mark.parametrize("value, exp_kind", (("1.2.0+dev1", "py"),))
    def test_init(self, value, exp_kind) -> None:
        assert exp_kind == Version(value).kind

    @pytest.mark.parametrize(
        "op, value, exp",
        (
            (operator.eq, "1.0.0", True),
            (operator.ge, "1.0.0", True),
            (operator.gt, "1.0.0", False),
            (operator.le, "1.0.0", True),
            (operator.lt, "1.0.0", False),
            (operator.ne, "1.0.0", False),
            # Not implemented
            pytest.param(
                operator.gt, 1.0, None, marks=pytest.mark.xfail(raises=TypeError)
            ),
        ),
    )
    def test_binop_str(self, op, value, exp) -> None:
        assert exp is op(Version("1.0.0"), value)


@pytest.mark.parametrize(
    "value, expected",
    (
        # SDMX 2.1
        ("0.0", PVVersion("0.0")),
        ("1.0", PVVersion("1.0")),
        # SDMX 3.0
        ("0.0.0-dev1", PVVersion("0.0.0+dev1")),
        ("1.0.0-dev1", PVVersion("1.0.0+dev1")),
        # Python
        ("1!1.2.3+abc.dev1", PVVersion("1!1.2.3+abc.dev1")),
        # Invalid
        pytest.param("foo", None, marks=pytest.mark.xfail(raises=InvalidVersion)),
    ),
)
def test_parse(value, expected) -> None:
    v = parse(value)

    assert expected == v

    # Value round-trips
    assert value == str(v)

    # Attributes can be accessed
    v.major
    v.minor
    v.patch
    v.local
    v.ext

    # Object's increment() method can be called
    assert v < v.increment(patch=1) < v.increment(minor=1) < v.increment(major=1)


@pytest.mark.parametrize(
    "kwargs, expected",
    (
        (dict(), PVVersion("1.1.0+dev1")),
        (dict(major=True), PVVersion("2.0.0")),
        (dict(major=1), PVVersion("2.0.0")),
        (dict(minor=True), PVVersion("1.1.0")),
        (dict(minor=1), PVVersion("1.1.0")),
        (dict(patch=True), PVVersion("1.0.1")),
        (dict(patch=1), PVVersion("1.0.1")),
        (dict(micro=True), PVVersion("1.0.1")),
        (dict(ext=1), PVVersion("1.0.0+dev1")),
        pytest.param(dict(foo=True), None, marks=pytest.mark.xfail(raises=ValueError)),
    ),
)
def test_increment0(kwargs, expected):
    # PVVersion.increment() method
    assert expected == parse("1.0.0").increment(**kwargs)

    # increment() function
    assert expected == increment("1.0.0", **kwargs)


_NIE = pytest.mark.xfail(raises=NotImplementedError)


@pytest.mark.parametrize(
    "base, kwarg, expected",
    (
        ("1.0.0", dict(ext=1), PVVersion("1.0.0+dev1")),
        ("1.0.0-dev1", dict(ext=1), PVVersion("1.0.0+dev2")),
        ("1.0.0-dev1", dict(ext=2), PVVersion("1.0.0+dev3")),
        ("1.0.0-foodev1", dict(ext=1), PVVersion("1.0.0+foodev2")),
        pytest.param("1.0.0-draft", dict(ext=1), None, marks=_NIE),
    ),
)
def test_increment1(base, kwarg, expected):
    """Test incrementing the 'extension' version part."""
    assert expected == parse(base).increment(**kwarg)



import re

import pytest

from sdmx.model import v21 as m
from sdmx.urn import expand, make, match, normalize, shorten


@pytest.mark.parametrize(
    "value, expected",
    (
        # MaintainableArtefact
        (
            "Codelist=BAZ:FOO(1.2.3)",
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
        ),
        # Item in a MaintainableArtefact
        (
            "Code=BAZ:FOO(1.2.3).BAR",
            "urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR",
        ),
        # Expand an already-complete URN: pass-through
        (
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
        ),
        # Not a URN: pass-through
        ("foo", "foo"),
    ),
)
def test_expand(value, expected) -> None:
    assert expected == expand(value)


def test_make() -> None:
    """:func:`.make` can look up and use information about the parent ItemScheme."""
    c = m.Code(id="BAR")

    with pytest.raises(
        ValueError, match="Neither <Code BAR> nor None are maintainable"
    ):
        make(c)

    cl = m.Codelist(id="FOO")
    cl.append(c)

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Cannot construct URN for <Codelist FOO (1 items)> without maintainer"
        ),
    ):
        make(c)

    cl.maintainer = m.Agency(id="BAZ")

    with pytest.raises(
        ValueError,
        match=re.escape(
            "Cannot construct URN for <Codelist BAZ:FOO (1 items)> without version"
        ),
    ):
        make(c, strict=True)

    cl.version = "1.2.3"

    assert "urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR" == make(c)
    assert "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)" == make(cl)


def test_match() -> None:
    # Value containing a "." in the ID
    urn = (
        "urn:sdmx:org.sdmx.infomodel.datastructure.Dataflow=LSD:"
        "TSTT1201R001_MIT1938_1_1._1(1.0)"
    )
    assert "TSTT1201R001_MIT1938_1_1._1" == match(urn)["id"]

    # Invalid value (only the "package" component, no "class")
    urn = "urn:sdmx:org.sdmx.infomodel.codelist=BBK:CLA_BBK_COLLECTION(1.0)"
    with pytest.raises(ValueError, match=re.escape(f"not a valid SDMX URN: {urn}")):
        match(urn)


@pytest.mark.parametrize(
    "value, expected",
    (
        # Other URN: pass-through
        (
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
        ),
        # Not a URN: pass-through
        ("foo", "foo"),
    ),
)
def test_normalize(value, expected) -> None:
    assert expected == normalize(value)


@pytest.mark.parametrize(
    "value, expected",
    (
        # MaintainableArtefact
        (
            "urn:sdmx:org.sdmx.infomodel.codelist.Codelist=BAZ:FOO(1.2.3)",
            "Codelist=BAZ:FOO(1.2.3)",
        ),
        # Item in a MaintainableArtefact
        (
            "urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR",
            "Code=BAZ:FOO(1.2.3).BAR",
        ),
        # Shorten an already-partial URN: pass-through
        ("Codelist=BAZ:FOO(1.2.3)", "Codelist=BAZ:FOO(1.2.3)"),
        # Not a URN: pass-through
        ("foo", "foo"),
    ),
)
def test_shorten(value, expected) -> None:
    assert expected == shorten(value)



"""Tests against the actual APIs for specific data sources."""

import logging
from pathlib import Path
from typing import Any, ClassVar, Union

import pytest

import sdmx
from sdmx import Client
from sdmx.exceptions import HTTPError, XMLParseError

log = logging.getLogger(__name__)


NI = "Not implemented in sdmx1"


class DataSourceTest:
    """Base class for data source tests."""

    # TODO also test structure-specific data

    #: Must be one of the IDs in sources.json.
    source_id: ClassVar[str]

    #: NB the following can be added to any subclass below for SSL failures. Update the
    #:    docstring to describe the nature of the problem.
    #: except skip SSL certificate verification
    client_args: ClassVar[dict] = {}

    #: Keyword arguments for particular endpoints.
    endpoint_args: ClassVar[dict[str, dict[str, Any]]] = {}

    #: True to xfail if a 503 Error is returned.
    tolerate_503: ClassVar[bool] = False

    #: Mapping of endpoint → Exception subclass. Tests of these endpoints are expected
    #: to fail with the given kind of exception.
    xfail: ClassVar[dict[str, Union[type[Exception], tuple[type[Exception], str]]]] = {}

    #: Failures affecting **all** data sources, internal to :mod:`sdmx`.
    xfail_common: ClassVar[dict[str, Any]] = {}

    @pytest.fixture(scope="class")
    def client(self, session_with_pytest_cache):
        """A Client that uses :func:`.session_with_pytest_cache`."""
        yield Client(self.source_id, session=session_with_pytest_cache)

    @pytest.fixture(scope="class")
    def client_with_stored_responses(self, session_with_stored_responses):
        """A Client that uses :func:`.session_with_stored_responses`."""
        yield Client(self.source_id, session=session_with_stored_responses)

    @pytest.mark.source
    @pytest.mark.network
    def test_endpoint(self, pytestconfig, client, endpoint, args):
        """Test all SDMX-REST endpoints for this data source.

        See also
        --------
        generate_endpoint_tests
        """
        message = client.get(endpoint, **args)

        if pytestconfig.getoption("verbose"):
            print(repr(message))
            if endpoint == "dataflow":
                # Display the IDs of data flows; this can be used to identify targets
                # for tests of the data endpoint for a particular source
                for dfd in message.dataflow:
                    print(repr(dfd))

        # For debugging
        # print(cache, cache.read_text(), result, sep="\n\n")
        # assert False

        # All parsed contents can also be converted to pandas
        sdmx.to_pandas(message)


@pytest.mark.usefixtures("testsource")
class TestTEST(DataSourceTest):
    source_id = "TEST"

    endpoint_args = {
        "schema": dict(context="datastructure"),
    }

    xfail = {
        "metadata": NotImplementedError,  # In .rest.v21.URL.handle_metadata()
        "registration": ValueError,  # In .rest.v21.URL.handle_registration()
    }

    # Same as above, but without the "source" or "network" marks
    def test_endpoint(self, client_with_stored_responses, endpoint, args):
        client_with_stored_responses.get(endpoint, **args)


class TestABS(DataSourceTest):
    source_id = "ABS"

    endpoint_args = {
        "data": dict(
            resource_id="ABS,ANA_AGG,1.0.0",
            key="....Q",
            params=dict(startPeriod="2020-Q1", endPeriod="2022-Q4"),
        )
    }


class TestABS_JSON(DataSourceTest):
    source_id = "ABS_JSON"

    endpoint_args = {
        "data": dict(
            resource_id="ABS,ANA_AGG,1.0.0",
            key="....Q",
        )
    }


class TestAR1(DataSourceTest):
    source_id = "AR1"

    endpoint_args = dict(
        data=dict(resource_id="WOE.XML"),
    )

    xfail = {
        "metadata": NotImplementedError,  # Internal to sdmx1
        "registration": ValueError,  # Internal to sdmx1
    }


class TestBBK(DataSourceTest):
    source_id = "BBK"

    # See https://github.com/khaeru/sdmx/issues/82
    xfail = {
        "codelist": XMLParseError,
        "datastructure": XMLParseError,
    }

    endpoint_args = {
        "data": dict(
            resource_id="BBDB2",
            key="H.DE.Y.A.C.IFRS.B.A.K.E.E001.VGH.A",
            params=dict(startPeriod="2017-S1", endPeriod="2019-S1"),
        ),
        "codelist": dict(
            resource_id="CL_BBK_ERX_RATE_TYPE", params=dict(references="none")
        ),
        "conceptscheme": dict(
            resource_id="CS_BBK_ERX",
            params=dict(detail="allstubs", references="parentsandsiblings"),
        ),
        "dataflow": dict(resource_id="BBEX3", params=dict(detail="referencestubs")),
        "datastructure": dict(resource_id="BBK_QFS"),
    }


class TestBIS(DataSourceTest):
    source_id = "BIS"

    endpoint_args = {
        "actualconstraint": dict(resource_id="CBP_D_24D"),
    }


class TestECB(DataSourceTest):
    source_id = "ECB"


# Data for mock responses; see TestESTAT.mock()
ESTAT_STORED = {
    "http://ec.europa.eu/eurostat/SDMX/diss-web/rest/data/nama_10_gdp/..B1GQ+P3.": {
        "content": Path("ESTAT", "footer2.xml"),
        "headers": {
            "Content-Type": "application/vnd.sdmx.genericdata+xml; version=2.1"
        },
    },
    "http://ec.europa.eu/eurostat/SDMX/diss-web/file/7JUdWyAy4fmjBSWT": {
        # This file is a trimmed version of the actual response for the above query
        "content": Path("ESTAT", "footer2.zip"),
        "headers": {"Content-Type": "application/octet-stream"},
    },
}


class TestESTAT(DataSourceTest):
    source_id = "ESTAT"

    @pytest.fixture
    def mock(self, client_with_stored_responses, test_data_path):
        from sdmx.util.requests import save_response

        for url, args in ESTAT_STORED.items():
            args["content"] = test_data_path.joinpath(args["content"]).read_bytes()
            save_response(client_with_stored_responses.session, "GET", url, **args)

        yield client_with_stored_responses

    def test_xml_footer(self, mock):
        msg = mock.get(url=list(ESTAT_STORED.keys())[0], get_footer_url=(1, 1))

        assert len(msg.data[0].obs) == 43

    @pytest.mark.network
    def test_ss_data(self, client):
        """Test a request for structure-specific data.

        Examples from:
        https://ec.europa.eu/eurostat/web/sdmx-web-services/example-queries
        """
        df_id = "NAMA_10_GDP"
        args = dict(resource_id=df_id)

        # Query for the DSD
        dsd = client.dataflow(**args).dataflow[df_id].structure

        # Even with ?references=all, ESTAT returns a short message with the DSD as an
        # external reference. Query again to get its actual contents.
        if dsd.is_external_reference:
            dsd = client.get(resource=dsd).structure[0]
            log.info(repr(dsd))

        assert not dsd.is_external_reference

        # Example query, using the DSD already retrieved
        args.update(
            dict(
                key=dict(unit=["CP_MEUR"], na_item=["B1GQ"], geo=["LU"]),
                params={"startPeriod": "2012", "endPeriod": "2015"},
                dsd=dsd,
                # commented: for debugging
                # tofile="temp.xml",
            )
        )
        client.data(**args)

    @pytest.mark.network
    def test_gh_116(self, caplog, client):
        """Test of https://github.com/khaeru/sdmx/issues/116.

        As of 2024-02-13, the actual web service no longer returns multiple versions of
        the same artefacts for this query.

        See also
        --------
        .test_reader_xml.test_gh_116
        """
        msg = client.get(
            "dataflow", "GOV_10Q_GGNFA", params=dict(detail="referencepartial")
        )

        if cl1 := msg.codelist.get("ESTAT:GEO(13.0)"):  # pragma: no cover
            # Both versions of the GEO codelist are accessible in the message
            cl2 = msg.codelist["ESTAT:GEO(13.1)"]

            # cl1 is complete and items are available
            assert not cl1.is_partial and 0 < len(cl1)
            # cl2 is partial, and fewer codes are included than in cl1
            assert cl2.is_partial and 0 < len(cl2) < len(cl1)
        else:
            assert msg.codelist["GEO"]

        if cl3 := msg.codelist.get("ESTAT:UNIT(15.1)"):  # pragma: no cover
            # Both versions of the UNIT codelist are accessible in the message
            cl4 = msg.codelist["ESTAT:UNIT(15.2)"]

            # cl3 is complete and items are available
            assert not cl3.is_partial and 0 < len(cl3)
            # cl4 is partial, and fewer codes are included than in cl1
            assert cl4.is_partial and 0 < len(cl4) < len(cl3)
        else:
            assert msg.codelist["UNIT"]


class TestESTAT3(DataSourceTest):
    source_id = "ESTAT3"

    # Examples from
    # https://wikis.ec.europa.eu/display/EUROSTATHELP/API+-+Getting+started+with+SDMX3.0+API
    endpoint_args = {
        "codelist": dict(resource_id="FREQ"),
        "conceptscheme": dict(resource_id="TESEM160"),
        "data": dict(
            context="dataflow", resource_id="ISOC_CI_ID_H", last_n_observations=100
        ),
        "dataflow": dict(resource_id="ISOC_CI_ID_H", version="1.0"),
        "datastructure": dict(resource_id="PRC_DAP13", params=dict(references="none")),
    }


class TestESTAT_COMEXT(DataSourceTest):
    source_id = "ESTAT_COMEXT"

    endpoint_args = {
        "data": dict(
            resource_id="DS-059271",
            params=dict(startPeriod="2023"),
        ),
    }


class TestCOMP(DataSourceTest):
    source_id = "COMP"

    endpoint_args = {
        "data": dict(resource_id="AID_RAIL"),
        "dataflow": dict(resource_id="AID_SCB_OBJ"),
    }


class TestEMPL(DataSourceTest):
    source_id = "EMPL"

    endpoint_args = {
        "data": dict(resource_id="LMP_IND_EXP"),
        "dataflow": dict(resource_id="LMP_IND_EXP"),
    }


class TestGROW(DataSourceTest):
    source_id = "GROW"

    endpoint_args = {
        "data": dict(resource_id="POST_CUBE1_X"),
        "dataflow": dict(resource_id="POST_CUBE1_X"),
    }


class TestILO(DataSourceTest):
    source_id = "ILO"

    endpoint_args = {
        "codelist": dict(resource_id="CL_ECO"),
    }

    xfail = {
        "agencyscheme": HTTPError,  # 400
        # TODO provide endpoint_args for the following 3 to select 1 or a few objects
        "contentconstraint": HTTPError,  # 413 Client Error: Payload Too Large
        "datastructure": HTTPError,  # 413 Client Error: Payload Too Large
        "organisationscheme": HTTPError,  # 400
        "structure": HTTPError,  # 400
        "structureset": NotImplementedError,  # 501
    }

    @pytest.mark.network
    def test_gh_96(self, caplog, client):
        """Test of https://github.com/khaeru/sdmx/issues/96.

        As of 2024-02-13, the web service no longer has the prior limitations on
        the `references` query parameter, so the special handling is removed.
        """
        client.get("codelist", "CL_ECO", params=dict(references="parentsandsiblings"))

        # As of 2024-02-13, no longer needed
        # assert (
        #     "ILO does not support references='parentsandsiblings'; discarded"
        #     in caplog.messages
        # )


class TestIMF(DataSourceTest):
    source_id = "IMF"


# As of 2025-01-10, all endpoints aside from SDMX 2.1 /data/ return 403
IMF_BETA_XFAIL: dict[str, Union[type[Exception], tuple[type[Exception], str]]] = {
    k: HTTPError
    for k in """
    actualconstraint
    agencyscheme
    allowedconstraint
    categorisation
    categoryscheme
    codelist
    conceptscheme
    contentconstraint
    dataconsumerscheme
    dataflow
    dataproviderscheme
    datastructure
    hierarchicalcodelist
    metadataflow
    metadatastructure
    organisationscheme
    provisionagreement
    registration
    structure
    structureset
    """.split()
}


class TestIMF_beta(DataSourceTest):
    source_id = "IMF_beta"

    endpoint_args = dict(
        # As indicated in the API documentation
        data=dict(
            resource_id="CPI",
            key="111.CPI.CP01.IX.M",
            params=dict(startPeriod=2018),
            # Does not appear to affect 403
            # headers={"User-Agent": "idata-script-client"},
        )
    )

    xfail = IMF_BETA_XFAIL | dict(
        metadata=NotImplementedError,
        registration=ValueError,
    )


class TestIMF_beta3(DataSourceTest):
    source_id = "IMF_beta3"

    endpoint_args = dict(
        data=dict(
            context="dataflow",
            agency_id="IMF",
            resource_id="CPI",
            key="111.CPI.CP01.IX.M",
            # Not yet supported
            # params={"c[TIME_PERIOD]": "ge:2018"},
        ),
        metadata=dict(provider_id="IMF"),
    )

    xfail = IMF_BETA_XFAIL | dict(
        data=HTTPError,  # 403
        metadata=HTTPError,  # 403
    )


class TestINEGI(DataSourceTest):
    source_id = "INEGI"

    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
        "structureset": NotImplementedError,  # 501
    }

    endpoint_args = dict(
        # 404 Not Found when the own source's ID ("INEGI") is used
        conceptscheme=dict(provider="ALL")
    )


class TestINSEE(DataSourceTest):
    source_id = "INSEE"

    xfail = {
        "agencyscheme": HTTPError,  # 400
        "contentconstraint": HTTPError,  # 400
        "organisationscheme": HTTPError,  # 400
        "structure": HTTPError,  # 400
        "structureset": HTTPError,  # 400
    }


class TestISTAT(DataSourceTest):
    source_id = "ISTAT"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }
    endpoint_args = {
        "actualconstraint": dict(resource_id="CONS_92_143"),
    }

    def test_gh_75(self, client_with_stored_responses):
        """Test of https://github.com/dr-leo/pandaSDMX/pull/75.

        - As of the original report on 2019-06-02, the 4th dimension was ``TIPO_DATO``,
          and the 5th ``TIPO_GESTIONE``.
        - As of 2021-01-30, these are transposed, and the 4th dimension name is
          ``TIPO_GEST``.
        - As of 2022-11-03, the dataflow uses yet another new DSD with the ID
          IT1:DCIS_SERVSOCEDU1(1.0).
        """
        client = client_with_stored_responses

        df_id = "47_850"

        # Reported Dataflow query works
        # Without references="datastructure", this is a very slow query
        df = client.dataflow(df_id, params={"references": "datastructure"}).dataflow[
            df_id
        ]

        # dict() key for the query
        data_key = dict(
            FREQ=["A"],
            REF_AREA=["001001+001002"],
            DATA_TYPE=["AUTP"],
            TYPE_SOCIO_EDUC_SERVICE=["ALL"],
            TYPE_MANAGEMENT=["ALL"],
            HOLDER_SECTOR_PUBPRIV=["1"],
        )

        # Dimension components are in the correct order
        assert [dim.id for dim in df.structure.dimensions.components] == list(
            data_key.keys()
        ) + ["TIME_PERIOD"]

        # Reported data query works
        # NB the reported query key was "A.001001+001002.1.AUTP.ALL.ALL"; adjusted per
        #    the DSD change (above).
        client.data(df_id, key="A.001001+001002.AUTP.ALL.ALL.1")

        # Use a dict() key to force Client to make a sub-query for the DSD
        try:
            client.data(df_id, key=data_key)
        except HTTPError as e:
            log.info(f"Known, sporadic {e!r}")
            pass

    @pytest.mark.xfail(
        raises=HTTPError,
        reason="""As of 2024-12-23, returns a 404 but also a text/plain message: «Error
        while retrieving Mappings from "Mapping Store"! Cause:Dataflow
        'urn:sdmx:org.sdmx.infomodel.datastructure.Dataflow=IT1:22_289(1.0)' doesn't
        contain a mapping set»""",
    )
    @pytest.mark.network
    def test_gh_104(self, client):
        """Test of https://github.com/khaeru/sdmx/issues/104.

        See also
        --------
        .test_reader_xml.test_gh_104
        """
        df_id = "22_289"

        dsd = (
            client.dataflow(df_id, params={"references": "datastructure"})
            .dataflow[df_id]
            .structure
        )

        # Data message is successfully parsed
        message = client.data(
            df_id,
            key=dict(AGE="TOTAL", SEX=["1", "2"], MARITAL_STATUS="99", REF_AREA="IT"),
            dsd=dsd,
            # tofile="debug.xml",
        )
        # Provided DSD is used to structure the data set(s) in the message
        assert dsd is message.data[0].structured_by


class TestLSD(DataSourceTest):
    source_id = "LSD"

    #: As of 2021-01-30, this source returns a certificate that is treated as invalid
    #: by the GitHub Actions job runner; but *not* on a local machine.
    client_args = dict(verify=False)

    endpoint_args = {
        # Using the example from the documentation
        "data": dict(
            resource_id="S3R629_M3010217",
            params=dict(startPeriod="2005-01", endPeriod="2007-01"),
        )
    }


class TestNB(DataSourceTest):
    """Norges Bank.

    This source returns a valid SDMX Error message (100 No Results Found) for the
    'categoryscheme' endpoint.
    """

    source_id = "NB"


class TestNBB(DataSourceTest):
    source_id = "NBB"
    endpoint_args = {
        "data": dict(
            resource_id="REGPOP",
            key="POPULA.000.",
            params=dict(startTime=2013, endTime=2017),
        )
    }


class TestOECD(DataSourceTest):
    source_id = "OECD"
    endpoint_args = {
        "actualconstraint": dict(resource_id="CR_A_DSD_DEBT_TRANS_COLL@DF_MICRO"),
        "data": dict(
            resource_id="DSD_MSTI@DF_MSTI",
            headers={"Accept-Encoding": "compress, gzip"},
        ),
    }


class TestOECD_JSON(DataSourceTest):
    source_id = "OECD_JSON"

    endpoint_args = {
        "data": dict(
            resource_id="ITF_GOODS_TRANSPORT",
            key=".T-CONT-RL-TEU+T-CONT-RL-TON",
            # Uncomment this to test https://github.com/khaeru/sdmx/issues/64; also
            # covered by a specimen in sdmx-test-data
            # resource_id="PART2",
        )
    }

    @pytest.fixture
    def client(self, cache_path):
        # Same as the default implementation, only using oecd_json.Client
        from sdmx.source.oecd_json import Client

        return Client(self.source_id, cache_name=str(cache_path), backend="sqlite")


class TestSGR(DataSourceTest):
    source_id = "SGR"


class TestSGR3(DataSourceTest):
    """Query the `SGR` source using SDMX 3.0."""

    source_id = "SGR"
    endpoint_args = {"codelist": dict(params=dict(format="sdmx-3.0"))}


class TestSPC(DataSourceTest):
    source_id = "SPC"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }
    endpoint_args = {
        "actualconstraint": dict(resource_id="CR_A_DF_ADBKI"),
        "data": dict(
            resource_id="DF_CPI",
            key="A.CK+FJ..",
            params=dict(startPeriod=2010, endPeriod=2015),
        ),
    }


class TestSTAT_EE(DataSourceTest):
    source_id = "STAT_EE"
    endpoint_args = {
        # Using the example from the documentation
        "data": dict(
            resource_id="VK12",
            key="TRD_VAL+TRD_VAL_PREV..TOTAL.A",
            params=dict(startTime=2013, endTime=2017),
        )
    }


class TestStatCan(DataSourceTest):
    source_id = "StatCan"

    endpoint_args = dict(
        data=dict(
            resource_id="DF_17100005",
            key=".1.138",
            params=dict(startPeriod=2015, endPeriod=2016),
        ),
        structure=dict(resource_id="Data_Structure_17100005"),
    )

    xfail = {
        "metadata": NotImplementedError,  # Internal to sdmx1
        "registration": ValueError,  # Internal to sdmx1
    }


class TestUNESCO(DataSourceTest):
    """UNESCO.

    Most endpoints are marked XFAIL because the service requires registration.
    """

    source_id = "UNESCO"
    xfail = {
        # Requires registration
        "categoryscheme": HTTPError,
        "codelist": HTTPError,
        "conceptscheme": HTTPError,
        "dataflow": HTTPError,
        "provisionagreement": HTTPError,
        # Because 'supports_series_keys_only' was set
        # TODO check
        # 'datastructure': NotImplementedError,
    }


class TestUNICEF(DataSourceTest):
    source_id = "UNICEF"

    @pytest.mark.network
    def test_data(self, client):
        dm = client.dataflow("GLOBAL_DATAFLOW")
        dsd = dm.structure[0]

        client.data("GLOBAL_DATAFLOW", key="ALB+DZA.MNCH_INSTDEL.", dsd=dsd)

        cl = dm.codelist["CL_UNICEF_INDICATOR"]
        c = cl["TRGT_2030_CME_MRM0"]

        # Code is properly associated with its parent, despite forward reference
        assert isinstance(c.parent, type(c))
        assert "TRGT_CME" == c.parent.id

    @pytest.mark.network
    def test_cd2030(self, client):
        """Test that :ref:`Countdown to 2030 <CD2030>` data can be queried."""
        dsd = client.dataflow("CONSOLIDATED", agency_id="CD2030").structure[0]

        # D5: Births
        client.data("CONSOLIDATED", key=dict(INDICATOR="D5"), dsd=dsd)


class TestUNSD(DataSourceTest):
    source_id = "UNSD"
    xfail = {
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
    }


class TestUY110(DataSourceTest):
    source_id = "UY110"

    xfail = {
        "metadata": NotImplementedError,  # Internal to sdmx1
        # 400: "Can not create reference, target structure is not maintainable, and no
        # identifiable reference parameters present"
        "organisationscheme": HTTPError,
        "registration": ValueError,  # Internal to sdmx1
        "structure": NotImplementedError,  # 501
    }


class TestWB(DataSourceTest):
    source_id = "WB"
    xfail = {
        "contentconstraint": NotImplementedError,  # 501
        "organisationscheme": HTTPError,  # 400
        "structure": NotImplementedError,  # 501
        "structureset": NotImplementedError,  # 501
    }


class TestWB_WDI(DataSourceTest):
    source_id = "WB_WDI"

    endpoint_args = {
        # Example from the documentation website
        "data": dict(
            resource_id="WDI",
            key="A.SP_POP_TOTL.AFG",
            params=dict(startPeriod="2011", endPeriod="2011"),
        )
    }



import re
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    import sdmx.model.common

#: Regular expression for URNs.
_PATTERN = re.compile(
    r"urn:sdmx:org\.sdmx\.infomodel"
    r"\.(?P<package>[^\.]*)"
    r"\.(?P<class>[^=]*)=((?P<agency>[^:]*):)?"
    r"(?P<id>[^\(]*)(\((?P<version>[\d\.]*)\))?"
    r"(\.(?P<item_id>.*))?"
)


class URN:
    """SDMX Uniform Resource Name (URN).

    For example: "urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR". The
    maintainer ID ("BAZ") and version ("1.2.3") must refer to a
    :class:`.MaintainableArtefact`. If (as in this example) the URN is for a
    non-maintainable child (for example, a :class:`.Item` in a :class:`.ItemScheme`),
    these are the maintainer ID and version of the containing scheme/other maintainable
    parent object.
    """

    #: SDMX :data:`.PACKAGE` corresponding to :attr:`klass`.
    package: str

    #: SDMX object class.
    klass: str

    #: ID of the :class:`.Agency` that is the :attr:`.MaintainableArtefact.maintainer`.
    agency: Optional[str] = None

    #: ID of the :class:`.MaintainableArtefact`.
    id: Optional[str] = None

    #: :attr:`.VersionableArtefact.version` of the maintainable artefact.parent.
    version: Optional[str] = None

    #: ID of an item within a maintainable parent. Optional.
    item_id: Optional[str] = None

    def __init__(self, value: Optional[str], **kwargs) -> None:
        if kwargs:
            self.__dict__.update(kwargs)

        if value is None:
            self.groupdict = {}  # Needed by match()
            return

        try:
            match = _PATTERN.match(value)
            assert match is not None
        except (AssertionError, TypeError):
            raise ValueError(f"not a valid SDMX URN: {value}")

        g = self.groupdict = match.groupdict()

        if g["package"] == "package":
            from sdmx.model.v21 import PACKAGE

            self.package = PACKAGE[g["class"]]
        else:
            self.package = g["package"]

        self.klass = g["class"]
        self.agency = g["agency"]
        self.id = g["id"]
        self.version = g["version"]
        self.item_id = g["item_id"]

    def __str__(self) -> str:
        return (
            f"urn:sdmx:org.sdmx.infomodel.{self.package}.{self.klass}={self.agency}:"
            f"{self.id}({self.version})"
            + (("." + self.item_id) if self.item_id else "")
        )


def expand(value: str) -> str:
    """Return the full URN for `value`.

    Example
    -------
    >>> expand("Code=BAZ:FOO(1.2.3).BAR")
    "urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR"

    Parameters
    ----------
    value : str
        Either the final / :func:`.shorten`'d part of a valid SDMX URN, or a full URN.

    Returns
    -------
    str
        The full SDMX URN. If `value` is not a partial or full URN, it is returned
        unmodified.
    """
    for candidate in (value, f"urn:sdmx:org.sdmx.infomodel.package.{value}"):
        try:
            return str(URN(candidate))
        except ValueError:
            continue

    return value


def make(
    obj,
    maintainable_parent: Optional["sdmx.model.common.MaintainableArtefact"] = None,
    strict: bool = False,
) -> str:
    """Create an SDMX URN for `obj`.

    If `obj` is not :class:`.MaintainableArtefact`, then `maintainable_parent` must be
    supplied in order to construct the URN.
    """
    from sdmx.model.common import MaintainableArtefact
    from sdmx.model.v21 import PACKAGE

    if not isinstance(obj, MaintainableArtefact):
        ma = maintainable_parent or obj.get_scheme()
        item_id = obj.id
    else:
        ma, item_id = obj, None

    if not isinstance(ma, MaintainableArtefact):
        raise ValueError(
            f"Neither {obj!r} nor {maintainable_parent!r} are maintainable"
        )
    elif ma.maintainer is None:
        raise ValueError(f"Cannot construct URN for {ma!r} without maintainer")
    elif strict and ma.version is None:
        raise ValueError(f"Cannot construct URN for {ma!r} without version")

    return str(
        URN(
            None,
            package=PACKAGE[obj.__class__.__name__],
            klass=obj.__class__.__name__,
            agency=ma.maintainer.id,
            id=ma.id,
            version=ma.version,
            item_id=item_id,
        )
    )


def match(value: str) -> dict[str, str]:
    """Match :data:`URN` in `value`, returning a :class:`dict` with the match groups.

    Example
    -------
    >>> match("urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR")
    {
        "package": "codelist",
        "class": "Code",
        "agency": "BAZ",
        "id": "FOO",
        "version": "1.2.3",
        "item_id": "BAR",
    }

    Raises
    ------
    ValueError
        If `value` is not a well-formed SDMX URN.
    """
    return URN(value).groupdict


def normalize(value: str) -> str:
    """‘Normalize’ a URN.

    Handle "…DataFlow=…" (SDMX 3.0) vs. "…DataFlowDefinition=…" (SDMX 2.1) in URNs;
    prefer the former.
    """
    return value.replace("Definition=", "=")


def shorten(value: str) -> str:
    """Return a partial URN based on `value`.

    Example
    -------
    >>> shorten("urn:sdmx:org.sdmx.infomodel.codelist.Code=BAZ:FOO(1.2.3).BAR")
    "Code=BAZ:FOO(1.2.3).BAR"

    Parameters
    ----------
    value : str
        A full SDMX URN. If `value` is not a URN, it is returned unmodified.

    Returns
    -------
    str
        `value`, but without the leading text
        :py:`"urn:sdmx:org.sdmx.infomodel.{package}."`
    """
    try:
        return str(URN(value)).split(".", maxsplit=4)[-1]
    except ValueError:
        return value



from . import Source as BaseSource


class Source(BaseSource):
    _id = "ISTAT"

    def modify_request_args(self, kwargs):
        """Supply explicit provider agency ID for ISTAT.

        As reported by @guglielmo, https://github.com/dr-leo/pandaSDMX/pull/75, ISTAT
        requires a provider agency ID for structure endpoints. Values "IT1" and "all"
        are known to work. Set "all" by default to (hopefully) be inclusive of "IT1" and
        any others.
        """
        super().modify_request_args(kwargs)

        # NB this is an indirect test for resource_type != 'data'; because of the way
        #    the hook is called, resource_type is not available directly.
        if "key" not in kwargs:
            kwargs.setdefault("agency_id", "all")



from pathlib import Path
from typing import TYPE_CHECKING, Any, Optional, Union
from warnings import warn

from . import csv, json, xml

if TYPE_CHECKING:
    import io
    from typing import TypeVar

    import sdmx.message
    import sdmx.reader.base

    T = TypeVar("T", bound=sdmx.reader.base.Converter)


#: All converters. Application code **may** extend this collection with custom
#: sub-classes of :class:`.Converter`.
CONVERTER = [csv.DataFrameConverter, csv.Reader, json.Reader, xml.Reader]

#: Only Readers for standard SDMX formats.
READERS = [csv.Reader, json.Reader, xml.Reader]


def detect_content_reader(content) -> type["sdmx.reader.base.BaseReader"]:
    """Return a reader class for :class:`bytes` `content`.

    .. deprecated:: 2.20.0
       Use :func:`get_reader` instead.
    """
    warn(
        "detect_content_reader(bytes); use get_reader() instead",
        DeprecationWarning,
        stacklevel=2,
    )
    return get_reader(content)


def _get(data: Any, kwargs: Optional[dict], _classes: list[type["T"]]) -> type["T"]:
    for c in _classes:
        if c.handles(data, kwargs or {}):
            return c

    raise ValueError(
        f"{data!r} not recognized by any of "
        + ", ".join(map(lambda c: c.__name__, _classes))
    )


def get_converter(
    data: Any, kwargs: Optional[dict] = None
) -> type["sdmx.reader.base.Converter"]:
    """Identify a :class:`Converter` or :class:`.Reader` for `data`.

    For each class in :data:`CONVERTER`, the :meth:`.Converter.handles` or
    :meth:`.BaseReader.handles` method is called with `data` and `kwargs`.

    `data` may include:

    - :class:`bytes` —same behaviour as deprecated :func:`.detect_content_reader`.
    - :class:`requests.Response` —same behaviour as deprecated
      :func:`.get_reader_for_media_type`.
    - :class:`pathlib.Path` —same behaviour as deprecated :func:`.get_reader_for_path`.

    …or, anything else that is handled by a class listed in :data:`CONVERTER`.

    Raises
    ------
    ValueError
        if none of the Converter classes can convert `data` and `kwargs` to SDMX.
    """
    return _get(data, kwargs, CONVERTER)


def get_reader(
    data: Any,
    kwargs: Optional[dict] = None,
    _classes: list[type["sdmx.reader.base.BaseReader"]] = READERS,
) -> type["sdmx.reader.base.BaseReader"]:
    """Identify a :class:`.Reader` for `data`.

    Identical to :func:`.get_converter`, except only :data:`READERS` for SDMX standard
    formats are returned.
    """
    return _get(data, kwargs, READERS)


def get_reader_for_media_type(value) -> type["sdmx.reader.base.BaseReader"]:
    """Return a reader class for HTTP content/media type `value`.

    .. deprecated:: 2.20.0
       Use :func:`get_reader` instead.
    """
    from requests import Response

    warn(
        "get_reader_for_media_type(str); use get_reader(requests.Response) instead",
        DeprecationWarning,
        stacklevel=2,
    )

    # Use `value` as Content-Type header for an otherwise-empty Response
    resp = Response()
    resp.headers["content-type"] = value

    try:
        return get_reader(resp)
    except ValueError as e:
        *_, names = e.args[0].partition(" any of ")
        raise ValueError(f"Media type {value!r} not supported by any of {names}")


def get_reader_for_path(path) -> type["sdmx.reader.base.BaseReader"]:
    """Return a reader class for file `path`.

    .. deprecated:: 2.20.0
       Use :func:`get_reader` instead.
    """
    warn(
        "get_reader_for_path(…); use get_reader() instead",
        DeprecationWarning,
        stacklevel=2,
    )

    p = Path(path)
    try:
        return get_reader(p)
    except ValueError as e:
        *_, names = e.args[0].partition(" any of ")
        raise ValueError(f"File suffix {p.suffix!r} not supported by any of {names}")


def read_sdmx(
    filename_or_obj: Union[bytes, str, Path, "io.IOBase", "io.BufferedReader"],
    format: Optional[str] = None,
    **kwargs,
) -> "sdmx.message.Message":
    """Read a :class:`.Message` from a path, file, or stream in an SDMX standard format.

    To identify whether `filename_or_obj` contains SDMX-CSV, SDMX-JSON, or SDMX-ML,
    :meth:`.BaseReader.handles` is called.

    Parameters
    ----------
    filename_or_obj :
        may include:

        - :class:`str` or :class:`pathlib.Path`: path to a particular file.
        - :class:`bytes`: raw/binary SDMX content.
        - :class:`io.IOBase`: a buffer, opened file, or other I/O object containing
          binary SDMX content.
    format : 'CSV', 'XML', or 'JSON', optional
        force handling `filename_or_obj` as if it had the given extension, even if
        :meth:`~.BaseReader.handles` fails to match.

    Other Parameters
    ----------------
    structure :
        :class:`.Structure`, :class:`.StructureUsage`, or other information used by a
        :class:`.BaseReader` to interpret the content of `filename_or_obj`. For example,
        the :class:`DataStructureDefinition <.BaseDataStructureDefinition>` for a
        structure-specific SDMX-ML message.
    """
    if isinstance(filename_or_obj, (str, Path)):
        path = Path(filename_or_obj)  # Ensure Path type
        obj: Union[bytes, "io.IOBase"] = open(path, "rb")  # Open the file
    else:
        path, obj = None, filename_or_obj

    # Try to identify a reader by first the path, then by the `obj` content
    for candidate in path, obj, Path(f"_.{(format or 'MISSING').lower()}"):
        try:
            reader = get_reader(candidate, kwargs)
        except ValueError:
            reader = None
        else:
            break

    if not reader:
        raise RuntimeError(
            f"cannot infer SDMX message format from path {path!r}, format "
            f"hint={format}, or content"
        )

    return reader().convert(obj, **kwargs)


def to_sdmx(data, **kwargs) -> "sdmx.message.Message":
    """Convert `data` in non-SDMX formats and data structures to SDMX :class:`.Message`.

    Unlike :func:`.read_sdmx`, which handles only the standard SDMX formats SDMX-CSV,
    SDMX-JSON, and SDMX-ML, this method can will process any Python data structure
    handled by a known :data:`CONVERTER`.
    """
    try:
        converter = get_converter(data, kwargs)
    except ValueError:
        raise NotImplementedError(f"Convert {type(data)} {data!r} to SDMX")

    return converter().convert(data, **kwargs)



from copy import copy
from typing import Iterable, Mapping, Optional, Sequence, Union

# TODO read this from the environment, or use any value set in the SDMX-ML spec.
#      Currently set to 'en' because test_dsd.py expects it.
#: Default locale for :class:`InternationalString`.
DEFAULT_LOCALE = "en"

# §3.2: Base structures


class InternationalString:
    """SDMX-IM InternationalString.

    SDMX-IM LocalisedString is not implemented. Instead, the 'localizations' is a
    mapping where:

    - keys correspond to the 'locale' property of LocalisedString.
    - values correspond to the 'label' property of LocalisedString.

    When used as a type hint with pydantic, InternationalString fields can be assigned
    to in one of four ways::

        @dataclass
        class Foo:
             name: InternationalStringDescriptor() = InternationalStringDescriptor()

        # Equivalent: no localizations
        f = Foo()
        f = Foo(name={})

        # Using an explicit locale
        f.name['en'] = "Foo's name in English"

        # Using a (locale, label) tuple
        f.name = ('fr', "Foo's name in French")

        # Using a dict
        f.name = {'en': "Replacement English name",
                  'fr': "Replacement French name"}

        # Using a bare string, implicitly for the DEFAULT_LOCALE
        f.name = "Name in DEFAULT_LOCALE language"

    Only the first method preserves existing localizations; the latter three replace
    them.

    """

    __slots__ = ("localizations",)

    # Types that can be converted into InternationalString
    _CONVERTIBLE = Union[str, Sequence, Mapping, Iterable[tuple[str, str]]]

    localizations: dict[str, str]

    def __init__(self, value: Optional[_CONVERTIBLE] = None, **kwargs):
        # Handle initial values according to type
        if value is None:
            # Keyword arguments → dict, possibly empty
            value = dict(kwargs)
        elif isinstance(value, str):
            # Bare string
            value = {DEFAULT_LOCALE: value}
        elif (
            isinstance(value, Sequence)
            and len(value) == 2
            and isinstance(value[0], str)
        ):
            # 2-tuple of str is (locale, label)
            value = {value[0]: value[1]}
        elif isinstance(value, Mapping):
            # dict; use directly
            value = dict(value)
        else:
            try:
                # Iterable of 2-tuples
                value = {  # type: ignore [misc]
                    locale: label for (locale, label) in value
                }
            except Exception:
                raise ValueError(value, kwargs)

        self.localizations = value

    # Convenience access
    def __getitem__(self, locale):
        return self.localizations[locale]

    def __setitem__(self, locale, label):
        self.localizations[locale] = label

    # Duplicate of __getitem__, to pass existing tests in test_dsd.py
    def __getattr__(self, name):
        try:
            return self.__getattribute__("localizations")[name]
        except KeyError:
            raise AttributeError(name) from None

    def __add__(self, other):
        result = copy(self)
        result.localizations.update(other.localizations)
        return result

    def localized_default(self, locale=None) -> str:
        """Return the string in *locale*, or else the first defined."""
        try:
            return self.localizations[locale]
        except KeyError:
            if len(self.localizations):
                # No label in the default locale; use the first stored value
                return next(iter(self.localizations.values()))
            else:
                return ""

    def __str__(self):
        return self.localized_default(DEFAULT_LOCALE)

    def __repr__(self):
        return "\n".join(
            ["{}: {}".format(*kv) for kv in sorted(self.localizations.items())]
        )

    def __eq__(self, other):
        try:
            return self.localizations == other.localizations
        except AttributeError:
            return NotImplemented


_TInternationalString = Union[InternationalString, InternationalString._CONVERTIBLE]
_TInternationalStringInit = Union[_TInternationalString, None]


class InternationalStringDescriptor:
    def __set_name__(self, owner, name):
        self._name = "_" + name

    def __get__(self, obj, type) -> InternationalString:
        if obj is None:
            return None  # type: ignore [return-value]

        return obj.__dict__[self._name]

    def __set__(self, obj, value: _TInternationalStringInit):
        if not isinstance(value, InternationalString):
            value = InternationalString(value)
        setattr(obj, self._name, value)



from . import Source as BaseSource


class Source(BaseSource):
    _id = "OECD"

    def modify_request_args(self, kwargs):
        """Supply explicit provider agency ID for OECD.

        The structures and data flows from this provider use a variety of agency IDs—for
        example “OECD.SDD.TPS”—to identify a specific the organizational unit within the
        OECD that is responsible for each object. Queries requesting structures or data
        with agency ID “OECD” (strictly) may return few or zero results.

        This hook sets the provider to “ALL” for structure queries if it is not given
        explicitly.
        """
        super().modify_request_args(kwargs)

        # NB this is an indirect test for resource_type != 'data'; because of the way
        #    the hook is called, resource_type is not available directly.
        if "key" not in kwargs:
            kwargs.setdefault("provider", "ALL")



from functools import lru_cache
from io import BytesIO, StringIO
from typing import TYPE_CHECKING

import pandas as pd
import pytest

import sdmx.message
from sdmx import to_sdmx
from sdmx.model import common
from sdmx.reader.csv import Handler, NotHandled, Reader

if TYPE_CHECKING:
    from sdmx.model import v30


class TestHandler:
    def test_abc(self):
        with pytest.raises(TypeError):
            Handler()

    def test_repr(self):
        assert "<NotHandled>" == repr(NotHandled())


class TestDataFrameConverter:
    @pytest.fixture
    def df(self) -> pd.DataFrame:
        """Equivalent to the content of v21/csv/example-01.csv"""
        buf = StringIO("""
STRUCTURE,STRUCTURE_ID,ACTION,DIM_1,DIM_2,DIM_3,OBS_VALUE,ATTR_2,ATTR_3,ATTR_1,UPDATED
dataflow,ESTAT:NA_MAIN(1.6.0),I,A,B,2014-01,12.4,Y,"Normal, special and other values",N,2021-01-22T13:15:41Z
dataflow,ESTAT:NA_MAIN(1.6.0),I,A,B,2014-02,10.8,Y,"Normal, special and other values",Y,2021-01-22T13:15:41Z
""")
        return pd.read_csv(buf)

    def test_to_sdmx(self, df) -> None:
        """:class:`.DataFrameConverter` can be used through :func:`.to_sdmx`."""
        dfd = get_dfd()

        result = to_sdmx(df, structure=dfd)  # Function runs

        assert isinstance(result, sdmx.message.DataMessage)  # Returns a data set
        assert 1 == len(result.data)  # Message has 1 data set
        ds = result.data[0]

        assert 2 == len(ds)  # Data set has 2 observations
        o0 = ds.obs[0]

        assert 12.4 == o0.value  # Observation has an expected value
        assert (  # Observation has the expected key
            dfd.structure.make_key(
                common.Key, dict(DIM_1="A", DIM_2="B", DIM_3="2014-01")
            )
            == o0.dimension
        )


class TestReader:
    @pytest.mark.parametrize(
        "mt, expected",
        [
            ("foo", False),
            ("application/vnd.sdmx.data+csv; version=1.0.0", True),
            ("application/vnd.sdmx.metadata+csv; version=2.0.0", True),
        ],
    )
    def test_handles_media_type(self, mt, expected) -> None:
        with pytest.warns(DeprecationWarning, match="use Converter.handles"):
            assert expected is Reader.handles_media_type(mt)

    @pytest.mark.parametrize(
        "content, exc_text",
        (
            (b"DATAFLOW,DIM_1,OBS_VALUE", "'DATAFLOW' in line 1, field 1"),
            (b"STRUCTURE,ACTION,DIM_1,OBS_VALUE", "'ACTION' in line 1, field 2"),
        ),
    )
    def test_inspect_header0(self, content, exc_text) -> None:
        with pytest.raises(ValueError, match=f"Invalid SDMX-CSV 2.0.0: {exc_text}"):
            Reader().convert(BytesIO(content))

    @pytest.mark.parametrize("value, expected", [(".csv", True), (".xlsx", False)])
    def test_supports_suffix(self, value, expected) -> None:
        with pytest.warns(DeprecationWarning, match="use Converter.handles"):
            assert expected is Reader.supports_suffix(value)


@lru_cache
def get_dfd(n_measure: int = 1) -> "v30.Dataflow":
    from sdmx.model import common, v30

    m = common.Agency(id="ESTAT")
    dsd = v30.DataStructureDefinition(maintainer=m)
    dsd.dimensions.append(v30.Dimension(id="DIM_1"))
    dsd.dimensions.append(v30.Dimension(id="DIM_2"))
    dsd.dimensions.append(v30.Dimension(id="DIM_3"))

    if n_measure == 1:
        dsd.measures.append(v30.Measure(id="OBS_VALUE"))
    else:
        for i in range(1, n_measure + 1):
            dsd.measures.append(v30.Measure(id=f"OBS_VALUE{i}"))

    dsd.attributes.append(common.DataAttribute(id="ATTR_2"))
    dsd.attributes.append(common.DataAttribute(id="ATTR_3"))
    dsd.attributes.append(common.DataAttribute(id="ATTR_1"))

    dfd = v30.Dataflow(id="NA_MAIN", maintainer=m, version="1.6.0", structure=dsd)

    return dfd


@pytest.mark.parametrize_specimens("path", format="csv")
def test_read_specimen(path) -> None:
    """Test that the samples from the SDMX-CSV spec can be read."""
    import sdmx

    if path.stem in ("example-02", "example-03"):
        dfd: "v30.Dataflow" = get_dfd(2)
    else:
        dfd = get_dfd()

    kwargs: dict = dict(structure=dfd)

    if path.stem == "example-04":
        kwargs.update(delimiter=";")

    result = sdmx.read_sdmx(path, **kwargs)

    # # DEBUG
    # from icecream import ic
    #
    # ic(result)
    # for i, ds in enumerate(result.data):
    #     try:
    #         ic(i, ds.action, sdmx.to_pandas(ds, attributes="dsgo"))
    #     except Exception as e:
    #         ic(e)
    #         pass

    del result



"""Tests for pandasdmx/writer.py."""

from typing import cast

import pandas as pd
import pandas.testing as pdt
import pytest
from pytest import raises

import sdmx
from sdmx.message import DataMessage, StructureMessage
from sdmx.model import common, v21
from sdmx.model.v21 import TimeDimension
from sdmx.testing import assert_pd_equal

MARKS = {
    "ESTAT/esms.xml": pytest.mark.xfail(raises=NotImplementedError),
}


def test_agencyscheme(specimen) -> None:
    # Convert an agency scheme
    with specimen("ECB/orgscheme.xml") as f:
        msg = sdmx.read_sdmx(f)
        data = sdmx.to_pandas(msg)

    assert data["organisation_scheme"]["SDMX:AGENCIES"]["ESTAT"] == "Eurostat"

    # to_pandas only returns keys for non-empty attributes of StructureMessage
    # https://github.com/dr-leo/pandaSDMX/issues/90
    assert set(data.keys()) == {"organisation_scheme"}

    # Attribute access works
    assert data.organisation_scheme["SDMX:AGENCIES"].ESTAT == "Eurostat"

    with pytest.raises(AttributeError):
        data.codelist
    with pytest.raises(AttributeError):
        data.dataflow
    with pytest.raises(AttributeError):
        data.structure


def test_categoryscheme(specimen) -> None:
    with specimen("IPI-2010-A21-structure.xml") as f:
        msg = sdmx.read_sdmx(f)
        data = sdmx.to_pandas(msg)

    cs = data["category_scheme"]["CLASSEMENT_DATAFLOWS"]

    assert cs.loc["COMPTA-NAT", "name"] == "National accounts (GDP, consumption...)"

    # Children appear
    assert cs.loc["CNA-PIB-2005", "parent"] == "CNA-PIB"


def test_codelist(specimen) -> None:
    # Retrieve codelists from a test specimen and convert to pandas
    with specimen("common-structure.xml") as f:
        dsd_common = sdmx.read_sdmx(f)
    codelists = sdmx.to_pandas(dsd_common)["codelist"]

    # File contains 5 code lists
    assert len(codelists) == 5

    # Code lists have expected number of items
    assert len(codelists["CL_FREQ"]) == 8

    # Items names can be retrieved by ID
    freq = codelists["CL_FREQ"]
    assert freq["A"] == "Annual"

    # Non-hierarchical code list has a string name
    assert freq.name == "Code list for Frequency (FREQ)"

    # Hierarchical code list
    with specimen("codelist_partial.xml") as f:
        msg = sdmx.read_sdmx(f)
        assert isinstance(msg, StructureMessage)

    # Convert single codelist
    CL_AREA = sdmx.to_pandas(msg.codelist["CL_AREA"])

    # Hierichical list has a 'parent' column; parent of Africa is the World
    assert CL_AREA.loc["002", "parent"] == "001"

    # Pandas features can be used to merge parent names
    area_hierarchy = pd.merge(
        CL_AREA,
        CL_AREA,
        how="left",
        left_on="parent",
        right_index=True,
        suffixes=("", "_parent"),
    )
    assert area_hierarchy.loc["002", "name_parent"] == "World"


def test_conceptscheme(specimen) -> None:
    with specimen("common-structure.xml") as f:
        msg = sdmx.read_sdmx(f)
        data = sdmx.to_pandas(msg)

    cdc = data["concept_scheme"]["CROSS_DOMAIN_CONCEPTS"]
    assert cdc.loc["UNIT_MEASURE", "name"] == "Unit of Measure"


@pytest.mark.parametrize_specimens("path", kind="data", marks=MARKS)
def test_data(specimen, path) -> None:
    if ("v3", "csv") == path.parts[-3:-1]:
        pytest.skip("SDMX-CSV 3.0.0 examples cannot be read without DSD")

    msg = sdmx.read_sdmx(path)

    result = sdmx.to_pandas(msg)

    expected = specimen.expected_data(path)
    if expected is not None:
        print(expected, result, sep="\n")
    assert_pd_equal(expected, result)

    # TODO incomplete
    assert isinstance(result, (pd.Series, pd.DataFrame, list)), type(result)


def test_data_decimal() -> None:
    """Test handling of "," as a decimal separator."""
    # Data set with string values containing "," as a decimal separator
    ds = v21.DataSet(
        obs=[
            v21.Observation(dimension=common.Key(FOO="A"), value="1"),
            v21.Observation(dimension=common.Key(FOO="B"), value="1,0"),
            v21.Observation(dimension=common.Key(FOO="C"), value="100,1"),
        ]
    )

    # Expected result
    exp = pd.Series(
        [1.0, 1.0, 100.1],
        index=pd.MultiIndex.from_product([list("ABC")], names=["FOO"]),
        name="value",
    )

    # Conversion occurs without error
    result = sdmx.to_pandas(ds)
    # Result is as expected
    pdt.assert_series_equal(exp, result)


def test_data_arguments(specimen) -> None:
    # The identity here is not important; any non-empty DataMessage will work
    with specimen("INSEE/CNA-2010-CONSO-SI-A17.xml") as f:
        msg = sdmx.read_sdmx(f)

    # Attributes must be a string
    with raises(TypeError):
        sdmx.to_pandas(msg, attributes=2)

    # Attributes must contain only 'dgso'
    with raises(ValueError):
        sdmx.to_pandas(msg, attributes="foobarbaz")


@pytest.mark.parametrize_specimens("path", kind="data", marks=MARKS)
def test_data_attributes(path) -> None:
    if ("v3", "csv") == path.parts[-3:-1]:
        pytest.skip("SDMX-CSV 3.0.0 examples cannot be read without DSD")

    msg = sdmx.read_sdmx(path)

    result = sdmx.to_pandas(msg, attributes="osgd")
    # TODO incomplete
    assert isinstance(result, (pd.Series, pd.DataFrame, list)), type(result)


def test_dataflow(specimen) -> None:
    # Read the INSEE dataflow definition
    with specimen("INSEE/dataflow") as f:
        msg = sdmx.read_sdmx(f)

    # Convert to pandas
    result = sdmx.to_pandas(msg, include="dataflow")

    # Number of Dataflows described in the file
    assert len(result["dataflow"]) == 663

    # ID and names of first Dataflows
    mbop = "Monthly Balance of Payments - "
    expected = pd.Series(
        {
            "ACT-TRIM-ANC": "Activity by sex and age - Quarterly series",
            "BPM6-CCAPITAL": "{}Capital account".format(mbop),
            "BPM6-CFINANCIER": "{}Financial account".format(mbop),
            "BPM6-CTRANSACTION": "{}Current transactions account".format(mbop),
            "BPM6-TOTAL": "{}Overall total and main headings".format(mbop),
        }
    )
    assert_pd_equal(result["dataflow"].head(), expected)


@pytest.mark.network
def test_dataset_constraint(specimen) -> None:
    """'constraint' argument to writer.write_dataset."""
    with specimen("ng-ts.xml") as f:
        msg = sdmx.read_sdmx(f)
        assert isinstance(msg, DataMessage)

    # Fetch the message's DSD
    assert msg.structure.is_external_reference
    # NB the specimen included in tests/data has 'ECB_EXR_NG' as the data structure ID;
    #    but a query against the web service gives 'ECB_EXR1' for the same data
    #    structure.
    id = "ECB_EXR1"
    dsd = cast(
        StructureMessage,
        sdmx.Client(msg.structure.maintainer.id).get("datastructure", id),
    ).structure[id]

    # Create a ContentConstraint
    cc = dsd.make_constraint({"CURRENCY": "JPY+USD"})

    # Write the message without constraint
    s1 = sdmx.to_pandas(msg)
    assert len(s1) == 12
    assert set(s1.index.to_frame()["CURRENCY"]) == {"CHF", "GBP", "JPY", "USD"}

    # Writing using constraint produces a fewer items; only those matching the
    # constraint
    s2 = sdmx.to_pandas(msg, constraint=cc)
    assert len(s2) == 6
    assert set(s2.index.to_frame()["CURRENCY"]) == {"JPY", "USD"}


def test_dataset_datetime(specimen) -> None:
    """Test datetime arguments to write_dataset()."""
    # Load structure
    with specimen("IPI-2010-A21-structure.xml") as f:
        dsd = cast(StructureMessage, sdmx.read_sdmx(f)).structure["IPI-2010-A21"]
        TIME_PERIOD = dsd.dimensions.get("TIME_PERIOD")
        FREQ = dsd.dimensions.get("FREQ")

    assert isinstance(TIME_PERIOD, TimeDimension)

    # Load data, two ways
    with specimen("IPI-2010-A21.xml") as f:
        msg = sdmx.read_sdmx(f, structure=dsd)
        assert isinstance(msg, DataMessage)
        ds = msg.data[0]
    with specimen("IPI-2010-A21.xml") as f:
        msg_no_structure = sdmx.read_sdmx(f)
        assert isinstance(msg_no_structure, DataMessage)

    other_dims = list(
        filter(lambda n: n != "TIME_PERIOD", [d.id for d in dsd.dimensions.components])
    )

    def expected(df, axis=0, cls=pd.DatetimeIndex):
        axes = ["index", "columns"] if axis else ["columns", "index"]
        assert getattr(df, axes[0]).names == other_dims
        assert isinstance(getattr(df, axes[1]), cls)

    # Write with datetime=str
    df = sdmx.to_pandas(ds, datetime="TIME_PERIOD")
    expected(df)

    # Write with datetime=Dimension instance
    df = sdmx.to_pandas(ds, datetime=TIME_PERIOD)
    expected(df)

    # Write with datetime=True fails because the data message contains no
    # actual structure information
    with pytest.raises(ValueError, match=r"no TimeDimension in \[.*\]"):
        sdmx.to_pandas(msg_no_structure, datetime=True)
    with pytest.raises(ValueError, match=r"no TimeDimension in \[.*\]"):
        sdmx.to_pandas(msg_no_structure.data[0], datetime=True)

    # DataMessage parsed with a DSD allows write_dataset to infer the
    # TimeDimension
    df = sdmx.to_pandas(msg, datetime=True)
    expected(df)
    # Same for DataSet
    df = sdmx.to_pandas(ds, datetime=True)
    expected(df)

    # As above, with axis=1
    df = sdmx.to_pandas(ds, datetime=dict(dim="TIME_PERIOD", axis=1))
    expected(df, axis=1)
    df = sdmx.to_pandas(ds, datetime=dict(dim=TIME_PERIOD, axis=1))
    expected(df, axis=1)
    ds.structured_by = dsd
    df = sdmx.to_pandas(ds, datetime=dict(axis=1))
    expected(df, axis=1)
    df = sdmx.to_pandas(msg, datetime=dict(axis=1))
    expected(df, axis=1)

    # Write with freq='M' works
    df = sdmx.to_pandas(ds, datetime=dict(dim="TIME_PERIOD", freq="M"))
    expected(df, cls=pd.PeriodIndex)

    # Write with freq='Y' (in older pandas, freq='A') works
    df = sdmx.to_pandas(ds, datetime=dict(dim="TIME_PERIOD", freq="Y"))
    expected(df, cls=pd.PeriodIndex)
    # …but the index is not unique, because month information was discarded
    assert not df.index.is_unique

    # Write specifying the FREQ dimension by name fails
    with pytest.raises(
        ValueError,
        match="cannot convert to PeriodIndex with " r"non-unique freq=\['A', 'M'\]",
    ):
        sdmx.to_pandas(ds, datetime=dict(dim="TIME_PERIOD", freq="FREQ"))

    # Remove non-monthly obs
    # TODO use a constraint, when this is supported
    ds.obs = list(filter(lambda o: o.key.FREQ != "A", ds.obs))

    # Now specifying the dimension by name works
    df = sdmx.to_pandas(ds, datetime=dict(dim="TIME_PERIOD", freq="FREQ"))

    # and FREQ is no longer in the columns index
    other_dims.pop(other_dims.index("FREQ"))
    expected(df, cls=pd.PeriodIndex)

    # Specifying a Dimension works
    df = sdmx.to_pandas(ds, datetime=dict(dim=TIME_PERIOD, freq=FREQ))
    expected(df, cls=pd.PeriodIndex)

    # As above, using DSD attached to the DataMessage
    df = sdmx.to_pandas(msg, datetime=dict(dim=TIME_PERIOD, freq="FREQ"))
    expected(df, cls=pd.PeriodIndex)

    # Invalid arguments
    with pytest.raises(ValueError, match="X"):
        sdmx.to_pandas(msg, datetime=dict(dim=TIME_PERIOD, freq="X"))
    with pytest.raises(ValueError, match="foo"):
        sdmx.to_pandas(ds, datetime=dict(foo="bar"))
    with pytest.raises(ValueError, match="43"):
        sdmx.to_pandas(ds, datetime=43)


def test_list_of_obs(specimen) -> None:
    """Bare list of observations can be written."""
    with specimen("ng-ts.xml") as f:
        msg = sdmx.read_sdmx(f)
        assert isinstance(msg, DataMessage)

    sdmx.to_pandas(msg.data[0].obs)


@pytest.mark.parametrize_specimens("path", kind="structure")
def test_structure(path) -> None:
    msg = sdmx.read_sdmx(path)

    sdmx.to_pandas(msg)

    # TODO test contents



import logging
from dataclasses import InitVar, dataclass, field
from enum import Enum, IntFlag
from functools import lru_cache
from typing import Literal, Optional, Union

from sdmx.util import parse_content_type

log = logging.getLogger(__name__)

#: Flag values for information about :class:`MediaType`:
#:
#: - ``data``: :obj:`True` if this format contains (meta)data. :obj:`False` if it
#:   contains (meta)data **structures**.
#: - ``meta``: :obj:`True` if this format contains metadata (or metadata structures).
#:   :obj:`False` otherwise.
#: - ``ss``: :obj:`True` if this format contains data that is structure-specific. This
#:   distinction is only relevant before SDMX 3.0.
#: - ``ts``: :obj:`True` if this format contains time-series data. This distinction is
#:   only relevant before SDMX 3.0.
Flag = IntFlag("Flag", "data meta ss ts")
f = Flag

#: SDMX standard versions. In this enumeration, the strings "3.0.0" and "3.0" evaluate
#: to the same member.
Version = Enum(
    "Version",
    {"1.0.0": 1, "2.0.0": 2, "2.1": 2.1, "3.0.0": 3, "3.0": 3, "unknown": None},
)


@dataclass(frozen=True)
class MediaType:
    """Structure of elements in :data:`MEDIA_TYPES`.

    The :func:`str` of a MediaType is generally of the form:

    ``application/vnd.sdmx.{label}+{base};version={version}``

    …unless :attr:`full` is provided, in which case label and base are ignored.
    """

    #: Distinguishing part of the media type.
    label: str

    #: The base media type or file format.
    base: Literal["csv", "json", "xml"]

    #: Format version.
    version: Version = field(init=False)
    _version: InitVar[Union[str, Version]]

    flags: Flag = Flag(0)

    #: Specify the full media type string.
    full: Optional[str] = None

    def __post_init__(self, _version):
        self.__dict__["version"] = Version[_version]

    def __repr__(self):
        """Full media type string."""
        return (
            self.full or f"application/vnd.sdmx.{self.label}+{self.base}"
        ) + f"; version={self.version.name}"

    @lru_cache()
    def match(self, value: str, strict: bool = False) -> bool:
        """:obj:`True` if `value` matches the current MediaType."""
        other = parse_content_type(value)
        other[1].pop("charset", None)
        this = parse_content_type(repr(self))

        if strict:
            return this == other
        else:
            result = this[0] == other[0]
            if result and this[1] != other[1]:
                log.debug(f"Match {this[0]} with params {other[1]}; expected {this[1]}")
            return result

    @property
    def is_data(self) -> bool:
        return bool(self.flags & Flag.data)

    @property
    def is_meta(self) -> bool:
        return bool(self.flags & Flag.meta)

    @property
    def is_structure_specific(self) -> bool:
        return bool(self.flags & Flag.ss)

    @property
    def is_time_series(self) -> bool:
        return bool(self.flags & Flag.ts)


#: SDMX formats. Each record is an instance of :class:`Format`.
MEDIA_TYPES = [
    # SDMX-ML
    # Version 2.1
    MediaType("generic", "xml", "2.1", f.data),
    MediaType("genericdata", "xml", "2.1", f.data),
    MediaType("genericmetadata", "xml", "2.1", f.data | f.meta),
    MediaType("generictimeseriesdata", "xml", "2.1", f.data | f.ts),
    MediaType("schema", "xml", "2.1", f.meta),
    MediaType("structure", "xml", "2.1"),
    MediaType("structurespecificdata", "xml", "2.1", f.data | f.ss),
    MediaType("structurespecificmetadata", "xml", "2.1", f.data | f.meta | f.ss),
    MediaType("structurespecifictimeseriesdata", "xml", "2.1", f.data | f.ss | f.ts),
    # Non-standard: returned by e.g. BIS, IMF, INSEE, LSD, NB, SGR, UNICEF
    MediaType("", "xml", "2.1", full="application/xml"),
    # Non-standard
    MediaType("", "xml", "2.1", full="text/xml"),
    # Version 3.0.0
    MediaType("data", "xml", "3.0.0", f.data),
    MediaType("structure", "xml", "3.0.0"),
    MediaType("metadata", "xml", "2.0.0", f.data),
    #
    # SDMX-JSON
    # Version 1.0.0
    MediaType("data", "json", "1.0.0", f.data),
    # Version 2.0.0
    MediaType("data", "json", "2.0.0", f.data),
    MediaType("structure", "json", "1.0.0"),
    MediaType("structure", "json", "2.0.0"),
    MediaType("metadata", "json", "2.0.0", f.data | f.meta),
    # Non-standard
    MediaType("draft-sdmx-json", "json", "1.0.0", f.data),
    # Non-standard; returned by e.g. NBB, STAT_EE
    MediaType("", "json", "1.0.0", f.data, full="draft-sdmx-json"),
    # Non-standard
    MediaType("", "json", "1.0.0", f.data, full="text/json"),
    #
    # SDMX-CSV
    # Version 1.0.0
    MediaType("data", "csv", "1.0.0", f.data),
    # Version 2.0.0
    MediaType("metadata", "csv", "2.0.0", f.data),
]


def list_media_types(**filters) -> list[MediaType]:
    """Return the string for each item in :data:`MEDIA_TYPES` matching `filters`."""
    result = []
    for mt in MEDIA_TYPES:
        if not all(getattr(mt, field) == value for field, value in filters.items()):
            continue
        result.append(mt)
    return result



"""Speed and memory usage tests."""

from sdmx.model.v21 import AttributeValue, DataAttribute, DataStructureDefinition


def test_refcount():
    # Component (subclasses) created outside of a DataStructureDefinition
    da1 = DataAttribute(id="foo")
    da2 = DataAttribute(id="foo")

    assert id(da1) != id(da2)
    assert da1 is not da2

    # Retrieving attributes from a DataStructureDefinition results in references
    # to the same object
    dsd = DataStructureDefinition()

    da3 = dsd.attributes.getdefault("foo")
    da4 = dsd.attributes.getdefault("foo")

    assert id(da3) == id(da4)
    assert da3 is da4
    assert len(dsd.attributes) == 1

    # Creating an AttributeValue referencing a DataAttribute outside a DSD
    av1 = AttributeValue(value="baz", value_for=DataAttribute(id="foo"))
    assert not any([av1.value_for is da for da in (da1, da2, da3)])

    # Same, using a DSD
    av2 = AttributeValue(value="baz", value_for="foo", dsd=dsd)
    assert av2.value_for is da3
